from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from logging import getLogger
from typing import Callable, Dict, Iterable, Iterator, NamedTuple, Optional, Union

from prosper_api.client import Client
from prosper_api.models import Listing, SearchListingsRequest

_BASE_REQUEST = SearchListingsRequest(
    limit=10,
    biddable=True,
    invested=False,
    sort_by="lender_yield",
    sort_dir="desc",
)

__all__ = [
    "AllocationStrategy",
    "AllocationStrategies",
    "FixedTargetAllocationStrategy",
    "HighestMatchingRateAllocationStrategy",
]

logger = getLogger(__file__)


class AllocationStrategy(Iterable[Listing]):
    """Defines a partial order over the set of active prosper listings.

    Specifically, it defines a sequence of prosper API calls to get listings and an optional sort method used to
    reorder them before emission. Use the timeout parameter to force a maximum time to wait for matching listings.
    """

    def __init__(
        self,
        client: Client,
        search_request_iterator: Iterator[SearchListingsRequest],
        local_sort: Optional[Callable[[Listing], bool]] = None,
        timeout_seconds: float = -1.0,
    ):
        """Gets an instance of AllocationStrategy.

        Args:
            client (Client): Prosper API client
            search_request_iterator (Iterator[SearchListingsRequest]): Iterates over the different parameters to pass
                the Listing api.
            local_sort (Optional[Callable[[Listing], bool]]): Sorts the returned listings into the desired final order
                before returning.
            timeout_seconds (float): The max real time in seconds to try to get a listing before giving up. `-1`
                (default) indicates no timeout is required.
        """
        self._client = client
        self._api_param_iterator = search_request_iterator
        self._local_sort = local_sort
        self._end_time = (
            datetime.now() + timedelta(seconds=timeout_seconds)
            if timeout_seconds > 0
            else None
        )
        # We hold on to this to help with debugging. It's not used after generation.
        self._search_request: Optional[SearchListingsRequest] = None
        self._buffer: Optional[Iterable[Listing]] = None

    def __next__(self) -> Listing:
        """Gets the next listing from the buffer or fetches a new one if needed and available.

        This method enforces the timeout set at instantiation.
        """
        if self._end_time is not None:
            if datetime.now() > self._end_time:
                raise StopIteration("Timed out while searching listings")

        if self._buffer is None:
            self._buffer = self._refresh_buffer()

        while True:
            try:
                next_val = next(self._buffer)
                break
            except StopIteration:
                # This might throw a StopIteration also; that means we're actually done
                self._buffer = self._refresh_buffer()

        return next_val

    def __iter__(self):
        """Implements the iterable interface."""
        return self

    def _refresh_buffer(self):
        self._search_request = self._api_param_iterator.__next__()
        if self._local_sort is not None:
            result = sorted(
                self._client.search_listings(self._search_request).result,
                key=self._local_sort,
            )
        else:
            result = self._client.search_listings(self._search_request).result

        return iter(result)


_AGGRESSIVE_TARGETS = {
    "HR": Decimal("0.02"),
    "E": Decimal("0.26"),
    "D": Decimal("0.23"),
    "C": Decimal("0.22"),
    "B": Decimal("0.15"),
    "A": Decimal("0.06"),
    "AA": Decimal("0.06"),
}

_CONSERVATIVE_TARGETS = {
    "HR": Decimal("0.02"),
    "E": Decimal("0.06"),
    "D": Decimal("0.06"),
    "C": Decimal("0.15"),
    "B": Decimal("0.22"),
    "A": Decimal("0.26"),
    "AA": Decimal("0.23"),
}


class _BucketDatum(NamedTuple):
    value: Union[float, Decimal]
    pct_of_total: Union[float, Decimal]
    error_pct: Union[float, Decimal]


class FixedTargetAllocationStrategy(AllocationStrategy):
    """Defines an investment strategy where funds are allocated to different prosper ratings at a fixed rate."""

    def __init__(
        self,
        client: Client,
        targets: Dict[str, Decimal],
    ):
        """Instantiates a FixedTargetAllocationStrategy.

        Args:
            client (Client): The prosper API client.
            targets (Dict[str, Decimal]): The target allocations by prosper rating.
        """
        buckets = {}
        account = client.get_account_info()
        invested_notes = account.invested_notes.model_dump()
        pending_bids = account.pending_bids.model_dump()
        total_account_value = account.total_account_value
        for rating in invested_notes.keys():
            # This assumes the ratings will never change
            value = invested_notes[rating] + pending_bids[rating]
            pct_of_total = value / total_account_value
            buckets[rating] = _BucketDatum(
                value=value,
                pct_of_total=pct_of_total,
                error_pct=(targets[rating] if rating in targets else 0) - pct_of_total,
            )

        buckets["Cash"] = _BucketDatum(
            account.available_cash_balance,
            account.available_cash_balance / total_account_value,
            0.0,
        )
        buckets["Pending deposit"] = _BucketDatum(
            account.pending_deposit,
            account.pending_deposit / total_account_value,
            0.0,
        )
        buckets["Total value"] = _BucketDatum(
            total_account_value, total_account_value / total_account_value, 0.0
        )

        logger.info(
            f"Pending investments = ${account.pending_investments_primary_market:8.2f}"
        )
        for key, bucket in buckets.items():
            logger.info(
                f"\t{key:17}= ${bucket.value:8.2f} ({bucket.pct_of_total * 100:6.2f}%) error: {bucket.error_pct * 100:6.3f}%"
            )

        grade_buckets_sorted_by_error_pct = sorted(
            (i for i in buckets.items() if i[0] in targets.keys()),
            key=lambda v: v[1].error_pct,
            reverse=True,
        )

        self._search_requests = [
            SearchListingsRequest(
                **{**_BASE_REQUEST.model_dump(), "prosper_rating": [b[0]]},
            )
            for b in grade_buckets_sorted_by_error_pct
        ]

        super().__init__(client, iter(self._search_requests))


class HighestMatchingRateAllocationStrategy(AllocationStrategy):
    """Allocation strategy that greedily takes the listing with the highest lender yield."""

    def __init__(self, client: Client, request: SearchListingsRequest):
        """Creates a new allocation strategy.

        Args:
            client (Client): Prosper client
            request (SearchListingsRequest): Base request for searching
        """
        self._search_requests = [request]
        super().__init__(client, iter(self._search_requests))


class AllocationStrategies(Enum):
    """Enumerates the pre-configured targets for AllocationStrategy."""

    AGGRESSIVE = (
        FixedTargetAllocationStrategy,
        _AGGRESSIVE_TARGETS,
    )
    CONSERVATIVE = (
        FixedTargetAllocationStrategy,
        _CONSERVATIVE_TARGETS,
    )
    OVERALL_HIGHEST_RATE = (
        HighestMatchingRateAllocationStrategy,
        _BASE_REQUEST,
    )

    def __str__(self):
        """Return the name of the enum to make it more palatable in the CLI help."""
        return self.name

    def to_strategy(self, client: Client) -> AllocationStrategy:
        """Converts the enum into a strategy, given a client and an account.

        Args:
            client (Client): Prosper client

        Returns:
            AllocationStrategy: Allocation strategy matching the inputs.
        """
        cls = self.value[0]
        args = self.value[1:]

        return cls(client, *args)



from prosper_shared.omni_config import Config, ConfigKey, input_schema

DRY_RUN_CONFIG = "prosper-bot.cli.dry-run"
VERBOSE_CONFIG = "prosper-bot.cli.verbose"


@input_schema
def _schema():
    return {
        "prosper-bot": {
            "cli": {
                ConfigKey(
                    "verbose", "Prints additional debug messages.", default=False
                ): bool,
                ConfigKey(
                    "dry-run",
                    "Run the loop but don't actually place any orders.",
                    default=False,
                ): bool,
            }
        }
    }


def build_config() -> Config:
    """Compiles all the config sources into a single config."""
    return Config.autoconfig(validate=True)



import logging
from numbers import Number
from typing import Any, Callable, Dict, Hashable, Iterable

import black

logger = logging.getLogger(__file__)


def repr_lambda(v: Any):
    """Intended to be able to get the raw representation of an object including the source of a lambda."""
    # if callable(v):
    #     return next(src_line for src_line in inspect.getsourcelines(v) if "lambda: " in src_line).replace(r".*lambda:", "lambda:")
    #
    # if isinstance(v, dict):
    #     return '{' + ', '.join(f"{k}: {repr_lambda(v)}" for k,v in v.items()) + '}'
    #
    # if isinstance(v, (list, tuple, set)):
    #     str_val = ', '.join(repr_lambda(v) for v in v)
    #     if isinstance(v, list):
    #         return '[' + str_val + ']'
    #     if isinstance(v, tuple):
    #         return '(' + str_val + ')'
    #     if isinstance(v, set):
    #         return '{' + str_val + '}'

    return repr(v)


def ppprint(o: Any) -> str:
    """Uses `black` to pretty-print Python objects.

    Arguments:
        o (Any): A raw Python value

    Returns:
        str: A formatted string representation of the value
    """
    return black.format_file_contents(repr_lambda(o), fast=False, mode=black.Mode())


def bucketize(
    input: Iterable[Any],
    bucketizer: Callable[[Any], Hashable] = lambda v: v,
    evaluator: Callable[[Any], Number] = lambda v: 1,
) -> Dict[Hashable, Number]:
    """Method to bucketize a series of values.

    Transforms a sequence of values into a dict of buckets where the bucket key is calculated using the provided `bucketizer`
    (the individual items by default) and the value summed is calculated using the provided `evaluator` (the count by default).
    The dict is sorted by the keys before returning.

    Arguments:
        input (Iterable[Any]): the collection of items to bucketize.
        bucketizer (Callable[[Any], Hashable]): A method to extract the key from the provided items. Defaults to return the item itself.
        evaluator (Callable[[Any], Number]): A method to extract the value from the provided items. Defaults to return a count of the items.

    Returns:
        Dict[Hashable, Number]: The result of summing the values into the calculated buckets.
    """
    unsorted_result = {}
    for item in input:
        bucket = bucketizer(item)
        value = evaluator(item)
        if bucket in unsorted_result:
            unsorted_result[bucket] += value
        else:
            unsorted_result[bucket] = value

    try:
        sorted_result = {k: unsorted_result[k] for k in sorted(unsorted_result.keys())}
    except TypeError:
        logger.warning("Sorting histogram by key failed; falling back to unsorted")
        sorted_result = unsorted_result

    return sorted_result


def print_histogram(
    title: str,
    histogram: Dict[Hashable, Number],
    percent: bool = True,
    printer: Callable[[str], object] = print,
):
    """Prints a string representation of a dict representing a histogram.

    Arguments:
        title (str): The title to print for the histogram.
        histogram (Dict[Hashable, Number]): The key/value pairs to render as a histogram.
        percent (bool): Whether the values should be converted to percentages of the total before rendering. Defaults to True.
        printer (Callable[[str], object]): The method to use to print, for example `logger.info`. Defaults to `print`.
    """
    max_len = max(len(str(key)) for key in histogram.keys())
    total = sum(histogram.values())
    printer(f"### {title} ###")
    for key, val in histogram.items():
        scaled_val = val / total * 100 if percent else val
        printer(
            f"{key:{max_len}}: {'#'*int(scaled_val)} {scaled_val:3.2f}{'%' if percent else ''}"
        )



import logging
from datetime import timedelta
from decimal import Decimal
from time import sleep
from typing import Union

import humanize
import simplejson as json
from prosper_api.client import Client
from prosper_shared.omni_config import ConfigKey, config_schema
from schema import Optional as SchemaOptional

from prosper_bot.allocation_strategy import AllocationStrategies
from prosper_bot.cli import DRY_RUN_CONFIG, VERBOSE_CONFIG, build_config

logger = logging.getLogger(__file__)

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s", level=logging.INFO
)

MIN_BID_CONFIG = "prosper-bot.bot.min_bid"
TARGET_LOAN_COUNT_CONFIG = "prosper-bot.bot.target-loan-count"
STRATEGY_CONFIG = "prosper-bot.bot.strategy"

POLL_TIME = timedelta(minutes=1)
MIN_ALLOWED_BID = Decimal("25.00")


@config_schema
def _schema():
    return {
        "prosper-bot": {
            "bot": {
                ConfigKey(
                    "min-bid",
                    "Minimum amount of a loan to purchase.",
                    default=MIN_ALLOWED_BID,
                ): Decimal,
                ConfigKey(
                    "strategy",
                    "Strategy for balancing your portfolio.",
                    default=AllocationStrategies.AGGRESSIVE,
                ): AllocationStrategies,
                SchemaOptional(
                    ConfigKey(
                        "target-loan-count",
                        "Calculate min bid based on (total account value / target loan count). Overrides min-bid if present.",
                    )
                ): int,
            }
        }
    }


class Bot:
    """Prosper trading bot."""

    strategy: AllocationStrategies

    def __init__(self, config=None):
        """Initializes the bot with the given argument values."""
        if config is None:
            config = build_config()
        self.config = config
        if self.config.get_as_bool(VERBOSE_CONFIG):
            logging.root.setLevel(logging.DEBUG)
            logger.setLevel(logging.DEBUG)

        self.client = Client(config=self.config)
        self.dry_run = self.config.get_as_bool(DRY_RUN_CONFIG)
        self.min_bid = self.config.get_as_decimal(MIN_BID_CONFIG, Decimal(25.00))
        self.target_loan_count = self.config.get(TARGET_LOAN_COUNT_CONFIG)
        self.strategy = self.config.get_as_enum(STRATEGY_CONFIG, AllocationStrategies)

    def run(self):
        """Main loop for the trading bot."""
        sleep_time_delta = POLL_TIME
        cash = None
        while True:
            try:
                cash, sleep_time_delta = self._do_run(cash)
            except KeyboardInterrupt:
                logger.info("Interrupted...")
                break
            except Exception as e:
                logger.warning(
                    f"Caught exception running bot loop: {e}. Continuing after {humanize.naturaldelta(sleep_time_delta)}..."
                )
                logger.debug("", exc_info=e)

            sleep(sleep_time_delta.total_seconds())

    def _do_run(self, previous_cash):
        account = self.client.get_account_info()
        logger.debug(json.dumps(account, indent=2, default=str))

        cash = account.available_cash_balance
        if previous_cash == cash:
            return cash, POLL_TIME

        allocation_strategy = self.strategy.to_strategy(self.client)

        invest_amount = self._get_bid_amount(
            cash, self.min_bid, account.total_account_value, self.target_loan_count
        )
        if invest_amount or self.dry_run:
            logger.info("Enough cash is available; searching for loans...")

            listing = next(allocation_strategy)
            lender_yield = listing.lender_yield
            listing_number = listing.listing_number
            if self.dry_run:
                logger.info(
                    f"DRYRUN: Would have purchased ${invest_amount:5.2f} of listing {listing_number} ({listing.prosper_rating}) at {lender_yield * 100:5.2f}% for {listing.listing_term} months"
                )
            else:
                order_result = self.client.order(listing_number, invest_amount)
                logging.info(
                    f"Purchased ${invest_amount:5.2f} of {listing_number} ({listing.prosper_rating}) at {lender_yield * 100:5.2f}% for {listing.listing_term} months"
                )
                logging.debug(json.dumps(order_result, indent=2, default=str))

            # Set the sleep time here in case of no matching listings being found (highly unlikely).
            sleep_time_delta = timedelta(seconds=5)
        else:
            sleep_time_delta = POLL_TIME
            logger.info(f"Starting polling once {humanize.naturaldelta(POLL_TIME)}...")

        return cash, sleep_time_delta

    @staticmethod
    def _get_bid_amount(
        cash: Decimal,
        min_bid: Decimal,
        total_account_value,
        target_loan_count: Union[int, None],
    ):
        if target_loan_count is not None:
            min_bid = max(
                _round_down_to_nearest_cent(total_account_value / target_loan_count),
                MIN_ALLOWED_BID,
            )

        logger.debug(f"Using ${min_bid} for min bid size")

        if cash < min_bid:
            return 0
        return _round_down_to_nearest_cent(min_bid + cash % min_bid)


def _round_down_to_nearest_cent(amount: Decimal):
    return amount.quantize(Decimal(".01"), rounding="ROUND_DOWN")


def runner():
    """Entry-point for Python script."""
    Bot().run()


if __name__ == "__main__":
    Bot().run()
